import click

from cogito.commands.scaffold_predict import scaffold_predict_classes
from cogito.core.config import (
    CogitoConfig,
    ConfigFile,
    FastAPIConfig,
    RouteConfig,
    ServerConfig,
    TrainingConfig,
)


def _init_with_default() -> ConfigFile:
    return ConfigFile.default()


def _init_prompted() -> ConfigFile:
    click.echo(
        "Please provide the following information to initialize the project configuration:"
    )
    name = click.prompt(
        "Project name", type=str, default="Cogito ergo sum", show_default=True
    )
    description = click.prompt(
        "Project description", type=str, default=None, show_default=True
    )
    version = click.prompt(
        "Project version", type=str, default="1.0.0", show_default=True
    )

    click.echo("Nice! Now let's configure the FastAPI settings:")

    host = click.prompt("Host", type=str, default="0.0.0.0", show_default=True)
    port = click.prompt("Port", type=int, default=8000, show_default=True)
    debug = click.confirm(
        "Would you like to run the API server in DEBUG mode?",
        default=False,
        show_default=True,
    )
    access_log = click.confirm(
        "Would you like to enable access logs?", default=False, show_default=True
    )

    fastapi = FastAPIConfig(
        host=host,
        port=port,
        debug=debug,
        access_log=access_log,
    )

    click.echo("This starts to look like an amazing inference service!")

    route = None

    if click.confirm(
        "Would you like to add a default route to the API?",
        default=True,
        show_default=True,
    ):
        route = RouteConfig.default()

    cache_dir = click.prompt(
        "Cache directory for model weights and artifacts",
        type=str,
        default="/tmp",
        show_default=True,
    )

    readiness_file = click.prompt(
        "Readiness file for health check",
        type=str,
        default="$HOME/readiness.lock",
        show_default=True,
    )

    server = ServerConfig(
        name=name,
        description=description,
        version=version,
        fastapi=fastapi,
        route=route,
        cache_dir=cache_dir,
        threads=1,
        readiness_file=readiness_file,
    )

    click.echo("Almost there! Let's configure the training settings.")

    # todo add training settings, when defined

    training = TrainingConfig()

    click.echo("Great! We're all set.")

    cogito = CogitoConfig(
        server=server,
        training=training,
    )

    return ConfigFile(cogito=cogito)


@click.command()
@click.option(
    "-s",
    "--scaffold",
    is_flag=True,
    default=False,
    help="Create a scaffold predict class in predict.py",
)
@click.option(
    "-d",
    "--default",
    is_flag=True,
    default=False,
    help="Initialize with default values",
)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    default=False,
    help="Force initialization, even if already initialized",
)
@click.pass_context
def init(
    ctx, scaffold: bool = False, default: bool = False, force: bool = False
) -> None:
    """Initialize the project configuration"""
    config_path = ctx.obj.get("config_path", ".") if ctx.obj else "."
    click.echo("Initializing...")

    if ConfigFile.exists(f"{config_path}/cogito.yaml") and not force:
        click.echo("Already initialized.")
        return

    if default:
        config = _init_with_default()
    else:
        config = _init_prompted()

    if scaffold:
        scaffold_predict_classes(config, force)

    config.save_to_file(f"{config_path}/cogito.yaml")
    click.echo("Initialized successfully.")



import asyncio
import logging
import os
import sys
from contextlib import asynccontextmanager
from typing import Any, Dict, Union

import uvicorn
from fastapi import FastAPI
from fastapi.exceptions import RequestValidationError

from cogito.api.handlers import (
    health_check_handler,
    metrics_handler,
)
from cogito.api.responses import ErrorResponse
from cogito.core.config import ConfigFile
from cogito.core.exceptioin_handlers import (
    too_many_requests_exception_handler,
    validation_exception_handler,
)
from cogito.core.exceptions import (
    ConfigFileNotFoundError,
    NoThreadsAvailableError,
    SetupError,
)
from cogito.core.logging import get_logger
from cogito.core.models import BasePredictor
from cogito.core.utils import (
    create_routes_semaphores,
    get_predictor_handler_return_type,
    load_predictor,
    wrap_handler,
    readiness_context,
)


class Application:
    _logger: logging.Logger
    ready: bool

    def __init__(
        self,
        config_file_path: str = ".",
        logger: Union[Any, logging.Logger] = None,
    ):

        self._logger = logger or Application._get_default_logger()

        try:
            self.config = ConfigFile.load_from_file(
                os.path.join(f"{config_file_path}/cogito.yaml")
            )
        except ConfigFileNotFoundError as e:
            self._logger.warning(
                "config file does not exist. Using default configuration.",
                extra={"error": str(e), "config_file_path": config_file_path},
            )
            self.config = ConfigFile.default()

        if self.config.cogito.server.cache_dir:
            os.environ["HF_HOME"] = self.config.cogito.server.cache_dir
            os.environ["COGITO_HOME"] = self.config.cogito.server.cache_dir
        else:
            os.environ["HF_HOME"] = os.path.expanduser("/.cogito/models")
            os.environ["COGITO_HOME"] = os.path.expanduser("/.cogito/models")

        @asynccontextmanager
        async def lifespan(app: FastAPI):

            try:
                await self.setup(app)
            except SetupError as e:
                self._logger.critical(
                    "Unable to start application",
                    extra={"error": e},
                )
                sys.exit(1)

            with readiness_context(self.config.cogito.server.readiness_file):
                yield

        self.app = FastAPI(
            title=self.config.cogito.server.name,
            version=self.config.cogito.server.version,
            description=self.config.cogito.server.description,
            access_log=self.config.cogito.server.fastapi.access_log,
            debug=self.config.cogito.server.fastapi.debug,
            lifespan=lifespan,
        )

        # FastAPIInstrumentor.instrument_app(self.app, excluded_urls=",".join(
        #        ["/health-check", "/metrics", "/docs", "/openapi.json"]))

        self.app.logger = self._logger

        self._set_default_routes()

        map_route_to_model: Dict[str, str] = {}
        self.map_model_to_instance: Dict[str, BasePredictor] = {}
        semaphores = create_routes_semaphores(self.config.cogito)

        route = self.config.cogito.server.route

        self._logger.info("Adding route", extra={"route": route})
        map_route_to_model[route.path] = route.predictor
        if route.predictor not in self.map_model_to_instance:
            predictor = load_predictor(route.predictor)
            self.map_model_to_instance[route.predictor] = predictor
        else:
            self._logger.info(
                "Predictor class already loaded",
                extra={"predictor": route.predictor},
            )

        model = self.map_model_to_instance.get(route.predictor)
        response_model = get_predictor_handler_return_type(model)

        handler = wrap_handler(
            descriptor=route.predictor,
            original_handler=getattr(
                self.map_model_to_instance.get(route.predictor), "predict"
            ),
            semaphore=semaphores[route.predictor],
            response_model=response_model,
        )

        self.app.add_api_route(
            route.path,
            handler,
            methods=["POST"],
            name=route.name,
            description=route.description,
            tags=route.tags,
            response_model=response_model,
            responses={500: {"model": ErrorResponse}},
        )

        self.app.add_exception_handler(
            RequestValidationError, validation_exception_handler
        )
        self.app.add_exception_handler(
            NoThreadsAvailableError, too_many_requests_exception_handler
        )

    def _set_default_routes(self) -> None:
        """Include default routes"""
        self.app.add_api_route(
            "/health-check",
            health_check_handler,
            methods=["GET"],
            name="health_check",
            description="Health check endpoint",
            tags=["health"],
        )

        self.app.add_api_route(
            "/metrics",
            metrics_handler,
            methods=["GET"],
            name="metrics",
            description="Metrics endpoint",
            tags=["metrics"],
        )

    async def setup(self, app: FastAPI):
        self._logger.info("Setting up application", extra={})
        for predictor in self.map_model_to_instance.values():
            try:
                self._logger.debug(
                    "Setting up predictor",
                    extra={"predictor": predictor.__class__.__name__},
                )
                # if is courutine
                if asyncio.iscoroutinefunction(predictor.setup):
                    await predictor.setup()
                else:
                    predictor.setup()
            except Exception as e:
                self._logger.critical(
                    "Unable to setting up predictor",
                    extra={"predictor": predictor.__class__.__name__, "error": e},
                )
                raise SetupError(predictor.__class__.__name__, e)

    def run(self):
        uvicorn.run(
            self.app,
            host=self.config.cogito.server.fastapi.host,
            port=self.config.cogito.server.fastapi.port,
        )

    @classmethod
    def _get_default_logger(cls):
        return get_logger("cogito.app")



from abc import ABC, abstractmethod
from typing import Any


class BasePredictor(ABC):
    @abstractmethod
    def predict(self, **kwargs) -> Any:
        pass

    @abstractmethod
    async def setup(self):
        pass



import asyncio
from contextlib import contextmanager
import importlib
import inspect
import logging
import os
import time
from inspect import Parameter, signature
from typing import Any, Callable, Dict, get_type_hints

try:
    # Pydantic v2
    from pydantic.fields import FieldInfo as Field
except ImportError:
    # Pydantic v1
    from pydantic.fields import Field

from pydantic import create_model

from cogito.api.responses import ErrorResponse, ResultResponse
from cogito.core.config import CogitoConfig
from cogito.core.exceptions import ModelDownloadError, NoThreadsAvailableError
from cogito.core.metrics import inference_duration_histogram
from cogito.core.model_store import download_gcp_model, download_huggingface_model
from cogito.core.models import BasePredictor


def load_predictor(class_path) -> Any:
    predictor_path, predictor_class = class_path.split(":")
    module = importlib.import_module(f"{predictor_path}")

    if not hasattr(module, predictor_class):
        raise AttributeError(
            f"Class {predictor_class} not found in module {predictor_path}"
        )

    predict_class = getattr(module, predictor_class)

    # Build an instance of the class
    predict_instance = predict_class()

    # Instantiate and return the class
    return predict_instance


def get_predictor_handler_return_type(predictor: BasePredictor):
    """This method returns the type of the output of the predictor.predict method"""
    # Get the return type of the predictor.predict method
    return_type = predictor.predict.__annotations__.get("return", Any)

    # Create a new dynamic type based on ResultResponse, with the correct module and annotated field
    return_class = type(
        f"{predictor.__class__.__name__}Response",
        (ResultResponse,),
        {
            "__annotations__": {
                "result": return_type
            },  # Annotate the result field with the return type
            "__module__": ResultResponse.__module__,  # Ensure the module is set correctly for Pydantic
        },
    )

    return return_class


def wrap_handler(
    descriptor: str,
    original_handler: Callable,
    response_model: ResultResponse,
    semaphore: asyncio.Semaphore = None,
) -> Callable:
    sig = signature(original_handler)
    type_hints = get_type_hints(original_handler)

    _, class_name = descriptor.split(":")

    input_fields = {}
    for name, param in sig.parameters.items():
        param_type = type_hints.get(name, Any)
        default_value = param.default if param.default != Parameter.empty else ...
        print(f"default_value: {default_value}")
        if not isinstance(default_value, type(...)):  # Comprueba si no es Ellipsis
            if isinstance(default_value, Field):
                default_value = default_value.default
        input_fields[name] = (param_type, Field(default=default_value))
    input_model = create_model(f"{class_name}Request", **input_fields)

    # Check if the original handler is an async function
    # Fixme Unify handler after replacing status checking model with file based mode.
    if inspect.iscoroutinefunction(original_handler):

        async def handler(input: input_model):
            async def a_timed_handler(input):
                result = None
                try:
                    dict_input = input.model_dump()
                except:
                    dict_input = input.dict()

                try:
                    start_time = time.time()
                    result = await original_handler(**dict_input)
                    end_time = time.time() - start_time
                    inference_duration_histogram.record(
                        end_time * 1000, {"predictor": class_name, "async": True}
                    )
                    # todo Count successful requests
                except Exception as e:
                    logging.exception(e)
                    # todo Count failed requests
                    return ErrorResponse(message=str(e)).to_json_response()

                return response_model(
                    inference_time_seconds=end_time,
                    input=dict_input,
                    result=result,
                )

            if not semaphore:
                return await a_timed_handler(input)
            else:
                if semaphore.locked():
                    raise NoThreadsAvailableError(descriptor)
                await semaphore.acquire()
                try:
                    return await a_timed_handler(input)
                finally:
                    semaphore.release()

    else:

        def handler(input: input_model):
            def timed_handler(intput: input_model):
                result = None
                try:
                    dict_input = input.model_dump()
                except:
                    dict_input = input.dict()
                try:
                    start_time = time.time()
                    result = original_handler(**dict_input)
                    end_time = time.time() - start_time
                    inference_duration_histogram.record(
                        end_time * 1000, {"predictor": class_name, "async": False}
                    )
                    # todo Count successful requests
                except Exception as e:
                    logging.exception(e)
                    # todo Count failed requests
                    return ErrorResponse(message=str(e)).to_json_response()

                return response_model(
                    inference_time_seconds=end_time,
                    input=dict_input,
                    result=result,
                )

            if not semaphore:
                return timed_handler(input)
            else:
                if semaphore.locked():
                    raise NoThreadsAvailableError(descriptor)
                semaphore.acquire()
                try:
                    return timed_handler(input)
                finally:
                    semaphore.release()

    handler.__annotations__ = {"input": input_model, "return": response_model}
    logging.debug(
        f"Handler of {original_handler.__name__} annotated with {handler.__annotations__}"
    )
    return handler


def model_download(model_path: str) -> str:
    """
    Download a model from various sources based on the model path format.
    Supported formats:
    - Google Cloud Storage: gs://bucket/path/to/model
    - Hugging Face: repo_owner/repo_name
    """
    cache_dir = os.getenv("COGITO_HOME")
    os.environ["HF_HOME"] = cache_dir

    try:
        if model_path.startswith("gs://"):
            return download_gcp_model(model_path, cache_dir)
        return download_huggingface_model(model_path, cache_dir)
    except Exception as e:
        raise ModelDownloadError(model_path, e)


def create_routes_semaphores(config: CogitoConfig) -> Dict[str, asyncio.Semaphore]:
    semaphores = {}
    route = config.server.route
    semaphores[route.predictor] = asyncio.Semaphore(config.server.threads)

    return semaphores


# Dependencia para limitar la concurrencia
async def limit_concurrent_requests(semaphore: asyncio.Semaphore):
    await semaphore.acquire()  # Bloquea si se alcanzó el límite

    try:
        yield  # Ejecuta la lógica de la ruta
    finally:
        semaphore.release()  # Libera el semáforo al finalizar


@contextmanager
def readiness_context(readiness_file: str) -> None:
    full_readiness_file = os.path.expandvars(os.path.expanduser(readiness_file))
    folder = os.path.dirname(full_readiness_file)
    os.makedirs(folder, exist_ok=True)

    with open(full_readiness_file, "w") as f:
        f.write("ready")
    yield
    os.remove(full_readiness_file)



import logging

import structlog

loggers = {}


# noinspection SpellCheckingInspection
class Logger(logging.Logger):
    def __init__(self, name, level=0):
        self._logger = structlog.get_logger(name)

    def setLevel(self, level):
        self._logger.setLevel(level)

    def isEnabledFor(self, level):
        return self._logger.isEnabledFor(level)

    def getEffectiveLevel(self):
        return self._logger.getEffectiveLevel()

    def getChild(self, suffix):
        return self._logger.getChild(suffix)

    def getChildren(self):
        return self._logger.getChildren()

    def debug(
        self, msg, *args, exc_info=None, stack_info=False, stacklevel=1, extra=None
    ):
        self._logger.debug(
            msg,
            *args,
            exc_info=exc_info,
            stack_info=stack_info,
            stacklevel=stacklevel,
            **extra,
        )

    def info(
        self, msg, *args, exc_info=None, stack_info=False, stacklevel=1, extra=None
    ):
        self._logger.info(
            msg,
            *args,
            exc_info=exc_info,
            stack_info=stack_info,
            stacklevel=stacklevel,
            **extra,
        )

    def warning(
        self, msg, *args, exc_info=None, stack_info=False, stacklevel=1, extra=None
    ):
        self._logger.warning(
            msg,
            *args,
            exc_info=exc_info,
            stack_info=stack_info,
            stacklevel=stacklevel,
            **extra,
        )

    def error(
        self, msg, *args, exc_info=None, stack_info=False, stacklevel=1, extra=None
    ):
        self._logger.error(
            msg,
            *args,
            exc_info=exc_info,
            stack_info=stack_info,
            stacklevel=stacklevel,
            **extra,
        )

    def exception(
        self, msg, *args, exc_info=True, stack_info=False, stacklevel=1, extra=None
    ):
        self._logger.exception(
            msg,
            *args,
            exc_info=exc_info,
            stack_info=stack_info,
            stacklevel=stacklevel,
            **extra,
        )

    def critical(
        self, msg, *args, exc_info=None, stack_info=False, stacklevel=1, extra=None
    ):
        self._logger.critical(
            msg,
            *args,
            exc_info=exc_info,
            stack_info=stack_info,
            stacklevel=stacklevel,
            **extra,
        )

    def log(
        self,
        level,
        msg,
        *args,
        exc_info=None,
        stack_info=False,
        stacklevel=1,
        extra=None,
    ):
        self._logger.log(
            level,
            msg,
            *args,
            exc_info=exc_info,
            stack_info=stack_info,
            stacklevel=stacklevel,
            **extra,
        )


def get_logger(name, level=0) -> Logger:
    if name not in loggers:
        loggers[name] = Logger(name, level)
    return loggers[name]



import os

import click
from jinja2 import Environment, FileSystemLoader

from cogito.core.config import ConfigFile
from cogito.core.exceptions import ConfigFileNotFoundError


def scaffold_predict_classes(config: ConfigFile, force: bool = False) -> None:
    template_dir = os.path.join(os.path.dirname(__file__), "..", "templates")
    env = Environment(loader=FileSystemLoader(template_dir))
    template = env.get_template("predict_class_template.jinja2")

    # TODO: validation of the config file?

    # Group routes by file
    files = {}
    route = config.cogito.server.route

    file_name = f'{route.predictor.split(":")[0]}.py'
    class_name = route.predictor.split(":")[1]
    class_data = route

    if file_name not in files:
        files[file_name] = []

    files[file_name].append({"class_name": class_name, "class_data": class_data})

    # Create the files
    for file, routes in files.items():
        class_names = ", ".join([route["class_name"] for route in routes])

        click.echo(f"Creating a scaffold predict classes ({class_names}) in {file}...")

        if os.path.exists(file) and not force:
            click.echo(f"File {file} already exists. Use --force to overwrite.")
            continue

        rendered_content = template.render(file=files, routes=routes)
        with open(file, "w") as f:
            f.write(rendered_content)

    pass


@click.command()
@click.option(
    "-f",
    "--force",
    is_flag=True,
    default=False,
    help="Force overwrite of existing files",
)
@click.pass_context
def scaffold(ctx, force: bool = False) -> None:
    """Generate predict classes"""

    config_path = ctx.obj.get("config_path", ".") if ctx.obj else "."

    click.echo("Generating predict classes...")

    # TODO: cogito.yaml file name should not be mandatory in the application
    try:
        config = ConfigFile.load_from_file(f"{config_path}/cogito.yaml")
    except ConfigFileNotFoundError:
        click.echo("No configuration file found. Please initialize the project first.")
        return

    scaffold_predict_classes(config, force)



import os
import sys

import click

from cogito import Application


@click.command()
@click.pass_obj
def run(ctx):
    """Run cogito app"""
    config_path = ctx.get("config_path")
    absolute_path = os.path.abspath(config_path)
    click.echo(f"Running '{absolute_path}' cogito application...")
    # change cwd to config_path
    os.chdir(absolute_path)
    if not os.path.exists(absolute_path):
        click.echo(
            f"Error: Path '{absolute_path}' does not exist.", err=True, color=True
        )
        exit(1)

    try:
        sys.path.insert(0, absolute_path)
        app = Application(config_file_path=absolute_path)
        app.run()
    except Exception as e:
        click.echo(f"Error: {e}", err=True, color=True)
        exit(1)
